/* libsocket++ - Socket abstraction library for C++
 *  Copyright (C) 2002  Simon Richter
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: builddSocket.h,v 1.3 2002/04/15 20:38:08 geier Exp $
 */

#ifndef _socketxx_Socket_h
#define _socketxx_Socket_h 1

#include <exception>
#include <string>

/* Included here since Solaris has defines for bind and connect, and since we
 * cannot avoid having it defined (or we won't find the functions when we need
 * them) the idea is to have the member functions renamed as well. As every-
 * one using our classes needs to include this header, it will be transparent
 * to the user. Having a real solution would be better IMO. */
#include <sys/socket.h>

namespace {

template<typename _CharT>
class Buffer {
	public:
		Buffer(void) : buffer(NULL), size(0), valid(0), offset(0) {}
		~Buffer(void) {
			if(buffer != NULL)
				free(buffer);
		}

	private:
		_CharT *buffer;
		unsigned int size;
		unsigned int valid;
		unsigned int offset;

	public:
		unsigned int getReadAvail(void) const {
			return valid - offset;
		}
		_CharT *getReadPosn(void) const {
			return &buffer[offset];
		}
		void advanceReadPosn(unsigned int num) {
			offset += num;
			if(offset == valid)
				offset = valid = 0;
		}

		unsigned int getWriteAvail(void) const {
			return size - valid;
		}

		_CharT *getWritePosn(void) const {
			return &buffer[valid];
		}

		void advanceWritePosn(unsigned int num) {
			valid += num;
		}

		void enlarge(void) {
			if(valid < size)
				return;
			if(offset > 0) {
				memcpy(buffer, &buffer[offset],
						(valid - offset) *
						sizeof(_CharT));
				valid -= offset;
				offset = 0;
				return;
			}
			_CharT *t = (_CharT *)realloc(buffer, size +
					(128 * sizeof(_CharT)));
			if(t != NULL) {
				buffer = t;
				size += 128;
			}
			return;
		}

		_CharT &operator[](unsigned int off) {
			return *(buffer + offset + off);
		}
};

}

namespace socketxx {

class Socket {
	protected:
		Socket(int fd);
	public:
		virtual ~Socket(void);

		static Socket *createSocket(const std::string &);

	private:
		int fd;
		Buffer<char> rdbuf;
		Buffer<char> wrbuf;

	public:
		int getfd(void) const { return fd; }

		void write(const void *data, size_t size);
		// This is just here for symmetry. You can avoid copying
		// the data around by using the subscription operator and
		// the advance() function below.
		size_t read(void *buffer, size_t size);

		virtual Socket *accept(void) = 0;
		virtual void bind(const std::string &) = 0;
		virtual void listen(int);
		virtual void connect(const std::string &) = 0;

		size_t available(void) { return rdbuf.getReadAvail(); }
		char &operator[](unsigned int off) { return rdbuf[off]; }
		void advance(unsigned int bytes) {
			rdbuf.advanceReadPosn(bytes);
		}

		bool needsWrite(void) {
			return (wrbuf.getReadAvail() > 0);
		}
		void flush(void);
		bool readData(void);

		Socket &operator<<(const std::string &s) {
			write(s.data(), s.length());
			return *this;
		}

	protected:
		virtual size_t sys_read(void *, size_t);
		virtual size_t sys_write(void *, size_t);

	protected:
		class Type {
			protected:
				Type(void);
				virtual ~Type(void);

			public:
				virtual const std::string &getType(void)
					const = 0;
				virtual Socket *operator()
					(const std::string &) const;
				virtual Socket *operator()
					(void) const = 0;

				static const Type &lookup(const std::string &);

			private:
				static Type *types[10];
				static int numTypes;
		};

	public:
		class InvalidAddressException : public std::exception {};
		class UnknownSocketTypeException : public std::exception {};
		class CannotCreateSocketException : public std::exception {};
		class CannotAcceptException : public std::exception {};
		class CannotBindException : public std::exception {};
		class CannotListenException : public std::exception {};
		class CannotConnectException : public std::exception {};
};

}

#endif
