/* libsocket++ - Socket abstraction library for C++
 *  Copyright (C) 2002  Simon Richter
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: builddSocket.cc,v 1.3 2002/04/15 20:38:09 geier Exp $
 */

#include <socket++/Socket.h>

#include <unistd.h>
#include <sys/socket.h>

namespace socketxx {

Socket::Socket(int fd) : fd(fd) {
	return;
}

Socket::~Socket(void) {
	::close(fd);
	return;
}

Socket *Socket::createSocket(const string &addr) {
	string::size_type pos = addr.find(' ');
	if(pos == string::npos)
		throw InvalidAddressException();
	string socktype(addr, 0, pos);
	string addrspec(addr, pos+1, string::npos);
	const Type &type = Type::lookup(socktype);
	return type(addrspec);
}

void Socket::write(const void *data, size_t size) {
	while(wrbuf.getWriteAvail() < size)
		wrbuf.enlarge();
	memcpy(wrbuf.getWritePosn(), data, size);
	wrbuf.advanceWritePosn(size);
	return;
}

size_t Socket::read(void *buffer, size_t size) {
	size_t available = rdbuf.getReadAvail();
	if(available > size)
		available = size;
	if(available <= 0)
		return 0;
	memcpy(buffer, rdbuf.getReadPosn(), available);
	return available;
}

void Socket::listen(int queue) {
	if(::listen(fd, queue) != 0)
		throw CannotListenException();
	return;
}

size_t Socket::sys_read(void *addr, size_t size) {
	return ::read(fd, addr, size);
}

size_t Socket::sys_write(void *addr, size_t size) {
	return ::write(fd, addr, size);
}

void Socket::flush(void) {
	if(wrbuf.getReadAvail() == 0)
		return;
	size_t bytes_written = sys_write(wrbuf.getReadPosn(),
			wrbuf.getReadAvail());
	if(bytes_written < 1)
		return;
	wrbuf.advanceReadPosn(bytes_written);
}

bool Socket::readData(void) {
	if(rdbuf.getWriteAvail() == 0)
		rdbuf.enlarge();
	size_t bytes_read = sys_read(rdbuf.getWritePosn(),
			rdbuf.getWriteAvail());
	if(bytes_read < 1)
		return false;
	rdbuf.advanceWritePosn(bytes_read);
	return true;
}

Socket::Type *Socket::Type::types[10];
int Socket::Type::numTypes = 0;

Socket::Type::Type(void) {
	types[numTypes++] = this;
	return;
}

Socket::Type::~Type(void) {
	for(int i = 0; i < numTypes; ++i) {
		if(types[i] != this)
			continue;
		for(; i < numTypes - 1; ++i) {
			types[i] = types[i+1];
		}
	}
	return;
}

const Socket::Type &Socket::Type::lookup(const string &type) {
	for(int i = 0; i < numTypes; ++i) {
		if(types[i]->getType() == type)
			return *types[i];
	}
	throw UnknownSocketTypeException();
}

Socket *Socket::Type::operator()(const string &addr) const {
	Socket *nsock = (*this)();
	nsock->bind(addr);
	return nsock;
}

}
